package game.engine.ai;

import java.util.Stack;

public abstract class AbstractPathFinder {

	public static final int START_OPEN = 0;

	public static final int DELETE_FROM_OPEN_LIST = 1;

	public static final int ADD_TO_CLOSED_LIST = 2;

	public static final int ADD_TO_OPEN_LIST = 3;

	private Node openListRoot = null;

	private Node closeListRoot = null;

	public Node currentNode = null;

	private Stack stack = new Stack();

	public int path_sx, path_sz, path_dx, path_dz, target_ID;

	public int rows;

	public Object data = null;

	public abstract int calculateCost(Node node, Node node2, int mode,
			Object data);

	public abstract boolean isChildValid(Node node, Node node2, int mode,
			Object data);

	public abstract void notifyChild(Node node, Node node2, int mode,
			Object data);

	public abstract void notifyList(Node node, Node node2, int mode, Object data);

	private final void clearNodes() {

		if (closeListRoot != null) {
			closeListRoot = null;
		}
		
		Node temp = null, temp2 = null;

		if (openListRoot != null) {
			while (openListRoot != null) {
				temp = openListRoot.next;

				openListRoot = null;
				openListRoot = temp;
			}
		}

		if (closeListRoot != null) {
			while (closeListRoot != null) {
				temp = closeListRoot.next;

				closeListRoot = null;
				closeListRoot = temp;
			}
		}
	}

	public final boolean generatePath(int sx, int sz, int dx, int dz) {
		this.stepInitialize(sx, sz, dx, dz);

		int retval = 0;

		while (retval == 0) {
			retval = this.step();
		}

		if (retval == -1 || this.currentNode == null) {
			this.currentNode = null;
			return false;
		}

		return true;
	}

	private final void stepInitialize(int sx, int sz, int dx, int dz) {

		this.clearNodes();

		path_sx = sx;
		path_sz = sz;
		path_dx = dx;
		path_dz = dz;

		target_ID = this.evaluateID(dx, dz);

		Node temp = new Node(sx, sz);

		temp.g = 0;
		temp.h = Math.abs(dx - sx) + Math.abs(dz - sz);
		temp.f = temp.g + temp.h;
		temp.nodeID = this.evaluateID(sx, sz);

		this.openListRoot = temp;

		this.notifyList(null, this.openListRoot, AbstractPathFinder.START_OPEN,
				data);
		this.notifyChild(null, temp, 0, data);
	}

	private final int step() {
		if ((this.currentNode = this.getBest()) == null) {
			return -1;
		}

		if (this.currentNode.nodeID == this.target_ID) {
			return 1;
		}

		this.createChildren(this.currentNode);

		return 0;
	}

	private final Node getBest() {

		if (this.openListRoot == null) {
			return null;
		}

		Node temp = this.openListRoot;
		Node temp2 = this.closeListRoot;

		this.openListRoot = temp.next;
		this.notifyList(null, temp, AbstractPathFinder.DELETE_FROM_OPEN_LIST,
				data);

		this.closeListRoot = temp;
		this.closeListRoot.next = temp2;

		this.notifyList(null, this.closeListRoot,
				AbstractPathFinder.ADD_TO_CLOSED_LIST, data);

		return temp;
	}

	private final void createChildren(Node node) {

		Node temp = new Node(-1, -1);
		int x = node.x;
		int z = node.z;

		temp.x = x;
		temp.z = z + 1;

		if (this.isChildValid(node, temp, 0, data)) {
			this.linkChild(node, temp);
		}

		temp.x = x + 1;
		temp.z = z;

		if (this.isChildValid(node, temp, 0, data)) {
			this.linkChild(node, temp);
		}

		temp.x = x;
		temp.z = z - 1;

		if (this.isChildValid(node, temp, 0, data)) {
			this.linkChild(node, temp);
		}

		temp.x = x - 1;
		temp.z = z;

		if (this.isChildValid(node, temp, 0, data)) {
			this.linkChild(node, temp);
		}
	}

	private final void linkChild(Node node, Node temp) {
		int x = temp.x;
		int z = temp.z;
		int g = node.g + this.calculateCost(node, temp, 0, data);
		int id = this.evaluateID(x, z);

		Node check = null;

		if ((check = this.checkList(this.openListRoot, id)) != null) {

			node.children.addElement(check);
			++node.childrenCount;

			if (g < check.g) {
				check.parent = node;
				check.g = g;
				check.f = g + check.h;

				this.notifyChild(node, check, 1, data);
			} else {
				this.notifyChild(node, check, 2, data);
			}
		} else if ((check = this.checkList(this.closeListRoot, id)) != null) {
			node.children.addElement(check);
			++node.childrenCount;

			if (g < check.g) {
				check.parent = node;
				check.g = g;
				check.f = g + check.h;

				this.notifyChild(node, check, 3, data);
				this.updateParents(check);
			} else {
				this.notifyChild(node, check, 4, data);
			}
		} else {
			Node newNode = new Node(x, z);
			newNode.parent = node;
			newNode.g = g;
			newNode.h = Math.abs(x - this.path_dx) + Math.abs(z - this.path_dz);
			newNode.f = newNode.g + newNode.h;
			newNode.nodeID = this.evaluateID(x, z);

			this.addToOpen(newNode);

			node.children.addElement(newNode);
			++node.childrenCount;

			this.notifyChild(node, newNode, 5, data);
		}
	}

	private final Node checkList(Node node, int id) {

		while (node != null) {
			if (node.nodeID == id) {
				return node;
			}

			node = node.next;
		}

		return null;
	}

	private final void addToOpen(Node addNode) {

		Node node = this.openListRoot;
		Node prev = null;

		if (this.openListRoot == null) {
			this.openListRoot = addNode;
			this.openListRoot.next = null;

			this.notifyList(null, addNode, AbstractPathFinder.START_OPEN, data);

			return;
		}

		while (node != null) {
			if (addNode.f > node.f) {
				prev = node;
				node = node.next;
			} else {
				if (prev != null) {
					prev.next = addNode;
					addNode.next = node;

					this.notifyList(prev, addNode,
							AbstractPathFinder.ADD_TO_OPEN_LIST, data);

				} else {
					Node temp = this.openListRoot;

					this.openListRoot = addNode;
					this.openListRoot.next = temp;

					this.notifyList(temp, addNode,
							AbstractPathFinder.START_OPEN, data);
				}

				return;
			}
		}

		prev.next = addNode;
		this.notifyList(prev, addNode, AbstractPathFinder.ADD_TO_OPEN_LIST,
				data);
	}

	private final void updateParents(Node node) {

		int g = node.g;
		int c = node.childrenCount;

		Node kid = null;

		for (int i = 0; i < c; i++) {
			kid = (Node) node.children.elementAt(i);

			if (g + 1 < kid.g) {

				kid.g = g + 1;
				kid.f = kid.g + kid.h;
				kid.parent = node;

				stack.push(kid);
			}
		}

		Node parent = null;

		while (!stack.empty()) {

			parent = (Node) stack.pop();
			c = parent.childrenCount;

			for (int i = 0; i < c; i++) {
				kid = (Node) parent.children.elementAt(i);

				if (parent.g + 1 < kid.g) {
					kid.g = parent.g + this.calculateCost(parent, kid, 0, data);
					kid.f = kid.g + kid.h;
					kid.parent = parent;

					stack.push(kid);
				}
			}
		}
	}

	private final int evaluateID(int x, int z) {
		return x * this.rows + z;
	}

}
