package game.engine.geometry;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.Vector;

import game.engine.GameflowEvent;
import game.engine.RouteMapElement;
import game.engine.renderer.Renderer;
import game.engine.renderer.TextureController;
import game.engine.spatial.LinearAlgebra4D;
import game.engine.spatial.Spatial;
import game.engine.spatial.Transformation;
import game.engine.*;

public abstract class Geometry extends Spatial {

	private int bufferID = -1;

	public static final int SIZEOF_FLOAT = 4;

	public static final int SIZEOF_INT = 4;

	public int VERTEX_COUNT = 0;

	public int NORMAL_COUNT = 0;

	public int INDEX_COUNT = 0;

	public int TEXTURE_COORD_COUNT = 0;

	public int TEXTURE_COORD_DIM = 0;

	public int DIMENSION = 0;

	public float[][] vertices = null;

	public float[][] normals = null;

	public float[][] textureCoord = null;

	public int[] indices = null;

	public FloatBuffer dataBuffer = null;

	public FloatBuffer vertexBuffer = null;

	public FloatBuffer normalBuffer = null;

	public FloatBuffer textureCoordinateBuffer = null;

	public IntBuffer indexBuffer = null;

	public int geometryType = -1;

	public Routemap routeMap = null;

	public boolean isSelectable = false;
	
	public boolean isShadow = false;
	
	public boolean isShadowPlane = false;
	
	public float[] ambientFactor = new float[4];
	
	public float[] diffuseFactor = new float[4];
	
	public float[] specularFactor = new float[4];
	
	public float shinessFactor = 0.0f;
	
	public float emmisionFactor = 0.0f; 
	
	private String textureID = null;

	public BoundingVolume boundingVolume = new BoundingVolume();

	public abstract void updateNormals();

	public abstract void calculateTexCoord();

	public abstract void executeStaticState();

	public abstract void executeDynamicState();

	public Geometry() {

	}

	public void setAmbientFactor(float[] material){
		this.ambientFactor[0] = material[0];
		this.ambientFactor[1] = material[1];
		this.ambientFactor[2] = material[2];
		this.ambientFactor[3] = material[3];
	}
	
	public void setDiffuseFactor(float[] material){
		this.diffuseFactor[0] = material[0];
		this.diffuseFactor[1] = material[1];
		this.diffuseFactor[2] = material[2];
		this.diffuseFactor[3] = material[3];
	}
	
	public void setSpecularFactor(float[] material){
		this.specularFactor[0] = material[0];
		this.specularFactor[1] = material[1];
		this.specularFactor[2] = material[2];
		this.specularFactor[3] = material[3];
	}
	
	public final int getGeometricType() {
		return this.geometryType;
	}

	public final void draw(Renderer renderer, boolean isCulled) {
		renderer.draw(this);
	}

	public void applyTransformation() {
		Transformation.concateTransformations(this.transformationMatrix,
				this.localTransformation, this.transformationMatrix);

		for (int i = 0; i < VERTEX_COUNT; i++) {
			LinearAlgebra4D.calculateTableVectorProduct(
					this.localTransformation.compositeMatrix, vertices[i],
					vertices[i]);
		}

		this.localTransformation.setIdentity();

		this.updateNormals();

		Geometry.updateBoundingSpere(this);
		Geometry.updateBuffers(this);
	}

	public final void updateTransformationState(
			Transformation newWorldTransformation) {
		Transformation.concateTransformations(newWorldTransformation,
				this.localTransformation, this.localTransformation);

		newWorldTransformation = null;

		this.applyTransformation();
	}

	public final void setTextureID(String textureID) {
		TextureController textureController = TextureController.getInstance();

		textureController.createTexture(textureID);
		this.textureID = textureID;
	}

	public final String getTextureID() {
		return this.textureID;
	}

	protected final static void createDataBuffer(Geometry geometry) {
		FloatBuffer vertexBuffer = newFloatBuffer(3 * geometry.VERTEX_COUNT
				* geometry.DIMENSION);

		geometry.dataBuffer = vertexBuffer;
	}

	protected final static void createDataBufferData(Geometry geometry) {
		geometry.dataBuffer.clear();

		geometry.dataBuffer.put(geometry.vertexBuffer);
		geometry.dataBuffer.put(geometry.normalBuffer);
		geometry.dataBuffer.put(geometry.textureCoordinateBuffer);
	}

	protected final static void createVertexBuffer(Geometry geometry) {

		FloatBuffer vertexBuffer = newFloatBuffer(geometry.VERTEX_COUNT
				* geometry.DIMENSION);

		geometry.vertexBuffer = vertexBuffer;
	}

	protected final static void createVertexBufferData(Geometry geometry) {
		geometry.vertexBuffer.clear();

		for (int i = 0; i < geometry.VERTEX_COUNT; i++) {
			geometry.vertexBuffer.put(geometry.vertices[i]);
		}
	}

	protected final static void createNormalBuffer(Geometry geometry) {
		FloatBuffer normalBuffer = newFloatBuffer(geometry.NORMAL_COUNT
				* geometry.DIMENSION - 1);

		geometry.normalBuffer = normalBuffer;
	}

	protected final static void createNormalBufferData(Geometry geometry) {
		geometry.normalBuffer.clear();

		for (int i = 0; i < geometry.NORMAL_COUNT; i++) {
			for (int j = 0; j < 3; j++) {
				geometry.normalBuffer.put(geometry.normals[i][j]);
			}

		}
	}

	protected final static void createTextureCoordBuffer(Geometry geometry) {
		FloatBuffer textureCoordinateBuffer = newFloatBuffer(geometry.TEXTURE_COORD_COUNT
				* geometry.TEXTURE_COORD_DIM);

		geometry.textureCoordinateBuffer = textureCoordinateBuffer;
	}

	protected final static void createTextureCoordinateBufferData(
			Geometry geometry) {
		geometry.textureCoordinateBuffer.clear();

		for (int i = 0; i < geometry.TEXTURE_COORD_COUNT; i++) {
			geometry.textureCoordinateBuffer.put(geometry.textureCoord[i]);
		}
	}

	public final static void updateBuffers(Geometry geometry) {
		Geometry.createVertexBufferData(geometry);
		Geometry.createNormalBufferData(geometry);
		Geometry.createDataBufferData(geometry);
	}

	private final static IntBuffer newIntBuffer(int numElements) {
		ByteBuffer bb = newByteBuffer(numElements * SIZEOF_INT);
		return bb.asIntBuffer();
	}

	private final static FloatBuffer newFloatBuffer(int numElements) {
		ByteBuffer bb = newByteBuffer(numElements * SIZEOF_FLOAT);
		return bb.asFloatBuffer();
	}

	private final static ByteBuffer newByteBuffer(int numElements) {
		ByteBuffer bb = ByteBuffer.allocateDirect(numElements);
		bb.order(ByteOrder.nativeOrder());
		return bb;
	}

	public int getBufferID() {
		return bufferID;
	}

	public void setBufferID(int[] bufferID) {
		// this.bufferID = bufferID[0];
	}

	private static final void calculateMostSeparatedBoundingPoints(Geometry geometry, int[] minMax) {

		int minX = 0;
		int maxX = 0;

		int minY = 0;
		int maxY = 0;

		int minZ = 0;
		int maxZ = 0;

		for (int i = 0; i < geometry.VERTEX_COUNT; i++) {

			if (geometry.vertices[i][0] < geometry.vertices[minX][0]) {
				minX = i;
			}

			if (geometry.vertices[i][0] > geometry.vertices[maxX][0]) {
				maxX = i;
			}

			if (geometry.vertices[i][1] < geometry.vertices[minY][1]) {
				minY = i;
			}

			if (geometry.vertices[i][1] > geometry.vertices[maxY][1]) {
				maxY = i;
			}

			if (geometry.vertices[i][2] < geometry.vertices[minZ][2]) {
				minZ = i;
			}

			if (geometry.vertices[i][2] > geometry.vertices[maxZ][2]) {
				maxZ = i;
			}
		}

		float[] temp1 = new float[4];
		float[] temp2 = new float[4];
		float[] temp3 = new float[4];

		LinearAlgebra4D.substractVectorVector(geometry.vertices[minX],
				geometry.vertices[maxX], temp1);
		LinearAlgebra4D.substractVectorVector(geometry.vertices[minY],
				geometry.vertices[maxY], temp2);
		LinearAlgebra4D.substractVectorVector(geometry.vertices[minZ],
				geometry.vertices[maxZ], temp3);

		float[] dist2X = new float[1];
		LinearAlgebra4D.calculateVectorDotProduct(temp1, temp1, dist2X);

		float[] dist2Y = new float[1];
		LinearAlgebra4D.calculateVectorDotProduct(temp2, temp2, dist2Y);

		float[] dist2Z = new float[1];
		LinearAlgebra4D.calculateVectorDotProduct(temp3, temp3, dist2Z);

		minMax[0] = minX;
		minMax[1] = maxX;

		if (dist2Y[0] > dist2X[0] && dist2Y[0] > dist2Z[0]) {
			minMax[0] = minY;
			minMax[1] = maxY;
		}

		if (dist2Z[0] > dist2X[0] && dist2Z[0] > dist2Z[0]) {
			minMax[0] = minZ;
			minMax[1] = maxZ;
		}
	}

	private static  final void boundingShereFromDistantPoints(Geometry geometry) {
		int[] minMax = new int[2];
		Geometry.calculateMostSeparatedBoundingPoints(geometry, minMax);

		LinearAlgebra4D.calculatePointVectorSum(geometry.vertices[minMax[0]],
				geometry.vertices[minMax[1]], geometry.boundingVolume.center);

		LinearAlgebra4D.scaleVector(geometry.boundingVolume.center, 0.5f,
				geometry.boundingVolume.center);
		
		float[] temp = new float[4];
		LinearAlgebra4D.substractVectorVector(geometry.boundingVolume.center,
				geometry.vertices[minMax[1]], temp);
		LinearAlgebra4D.calculateVectorDotProduct(temp, temp,
				geometry.boundingVolume.radius);

		geometry.boundingVolume.radius[0] = (float) Math
				.sqrt((double) geometry.boundingVolume.radius[0]);

	}

	private static final void boundingSphereOfSphereAndPoints(Geometry geometry, float[] vertex) {

		float[] temp = new float[4];
		LinearAlgebra4D.substractVectorVector(geometry.boundingVolume.center,
				vertex, temp);

		float dist[] = new float[1];
		LinearAlgebra4D.calculateVectorDotProduct(temp, temp, dist);

		if (dist[0] > geometry.boundingVolume.radius[0]
				* geometry.boundingVolume.radius[0]) {

			float d = (float) Math.sqrt((double) dist[0]);
			float newRadius = (geometry.boundingVolume.radius[0] + d) * 0.5f;
			float k = (newRadius - geometry.boundingVolume.radius[0]) / d;

			geometry.boundingVolume.radius[0] = newRadius;

			LinearAlgebra4D.scaleVector(temp, k, temp);
			LinearAlgebra4D.calculatePointVectorSum(geometry.boundingVolume.center,
					temp, geometry.boundingVolume.center);
		}

	}

	public static void updateBoundingSpere(Geometry geometry) {
		Geometry.boundingShereFromDistantPoints(geometry);

		for (int i = 0; i < geometry.VERTEX_COUNT; i++) {
			Geometry.boundingSphereOfSphereAndPoints(geometry, geometry.vertices[i]);
		}
	}

}
