package game.engine.renderer;

import game.engine.Node;
import game.engine.RouteMapElement;
import game.engine.Routemap;
import game.engine.SceneGraph;
import game.engine.SelectionController;
import game.engine.geometry.Geometry;
import game.engine.spatial.Spatial;
import game.engine.spatial.Transformation;

import java.awt.Component;
import java.nio.Buffer;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.IntBuffer;

import net.java.games.jogl.GL;
import net.java.games.jogl.GLDrawable;
import net.java.games.jogl.GLU;
import net.java.games.jogl.util.GLUT;

public class OpenGLRenderer extends Renderer {

	public GLDrawable drawable = null;

	public GL gl = null;

	public GLU glu = null;

	public float[] light_position = { 20f, 40f, 20f, 0f };

	public float[] light_diffuse = {  0.4f, 0.4f, 0.4f, 1.0f  };

	public float[] light_ambient = {  0.2f, 0.2f, 0.2f, 1.0f  };

	public float[] light_specular = { 1f, 1f, 1f, 1.0f };

	private static final int SELECTION_BUFFER_LENGTH = 128;

	public static final int SIZEOF_INT = 4;

	private int mouseX = -1;

	private int mouseY = -1;

	private IntBuffer selectionBuffer = OpenGLRenderer
			.newIntBuffer(OpenGLRenderer.SELECTION_BUFFER_LENGTH);

	public OpenGLRenderer(GLDrawable drawable) {
		this.drawable = drawable;
		this.gl = drawable.getGL();
		this.glu = drawable.getGLU();

		this.setWidth();
		this.setHeight();

		this.sceneGraph = new SceneGraph(this);
	}

	public Object getRenderingDevice() {
		return this.gl;
	}

	public final void enableDepthTest() {
		gl.glEnable(GL.GL_DEPTH_TEST);

		gl.glEnable(GL.GL_CULL_FACE);
		gl.glCullFace(GL.GL_BACK);
		gl.glFrontFace(GL.GL_CCW);
	}

	public final void enableLight() {
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, light_position);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, light_ambient);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, light_diffuse);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, light_specular);
		gl.glLightModeli(GL.GL_LIGHT_MODEL_COLOR_CONTROL, GL.GL_SEPARATE_SPECULAR_COLOR);
		
		gl.glEnable(GL.GL_LIGHTING);
		gl.glEnable(GL.GL_LIGHT0);
		gl.glEnable(GL.GL_COLOR_SUM);
		
		gl.glShadeModel(GL.GL_SMOOTH);
	}

	public final void initialBufferObjects() {
		/*String version = gl.glGetString(GL.GL_VERSION);

		if ((version.charAt(0) == '1') && (version.charAt(1) == '.')
				&& (version.charAt(2) >= '5') && (version.charAt(2) <= '9')) {
			System.out.println("GL.OK");
			System.out.println(version);
		} else {
			System.out.println("false");
			System.out.println(version);

		}*/
	}

	public final void enableVertexArrays() {
		gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
		gl.glEnableClientState(GL.GL_NORMAL_ARRAY);
		gl.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY);
	}

	public final void enableFog() {
		float[] fogColor = { 0.2f, 0.2f, 0.2f, 1.0f };
		
		gl.glEnable(GL.GL_FOG);
		gl.glFogi(GL.GL_FOG_MODE, GL.GL_EXP2);
		gl.glFogfv(GL.GL_FOG_COLOR, fogColor);
		gl.glFogf(GL.GL_FOG_DENSITY, 0.02f);
		gl.glFogf(GL.GL_FOG_START, 50.0f);
		gl.glFogf(GL.GL_FOG_END, 60.0f);
	}

	public final void enableAntialiasing() {
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glEnable(GL.GL_BLEND);
		gl.glEnable(GL.GL_LINE_SMOOTH);
		gl.glHint(GL.GL_LINE_SMOOTH_HINT, GL.GL_NICEST);
	}

	public final void enableTextures() {
		gl.glEnable(GL.GL_TEXTURE_2D);
		gl.glTexEnvi(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_MODULATE);

	}

	public final void initialProjection() {
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();

		int w = this.width;
		int h = this.height;

		double aspect = w / h;

		glu.gluPerspective(65.0, aspect, 2.0, 60.0);
	}

	public final void initialDrawing() {
		// gl.glBegin(GL.GL_TRIANGLES);
	}

	@Override
	public final boolean beginScene() {
		if (this.renderMode == Renderer.RENDER_MODE) {
			this.sceneGraph.notifyGameflowController();
			this.clearBuffers();
			
			gl.glMatrixMode(GL.GL_MODELVIEW);
			gl.glLoadIdentity();

			float[] eyePoint = this.camera.eyePoint;
			float[] lookAtPoint = this.camera.lookAtPoint;
			float[] upVector = this.camera.upVector;

			glu.gluLookAt(eyePoint[0], eyePoint[1], eyePoint[2], lookAtPoint[0],
					lookAtPoint[1], lookAtPoint[2], upVector[0], upVector[1],
					upVector[2]);

			ViewFrustrum.getInstance().updateFrustrum();
		}

		if (this.renderMode == Renderer.SELECTION_MODE) {
			
			
			int[] viewPort = new int[4];
			float fAspect = 0.0f;

			this.selectionBuffer.clear();
			gl.glSelectBuffer(OpenGLRenderer.SELECTION_BUFFER_LENGTH,
					this.selectionBuffer);
			gl.glInitNames();
			gl.glPushName(0);

			gl.glGetIntegerv(GL.GL_VIEWPORT, viewPort);

			gl.glMatrixMode(GL.GL_PROJECTION);
			gl.glPushMatrix();

			gl.glRenderMode(GL.GL_SELECT);
			gl.glLoadIdentity();

			glu.gluPickMatrix(this.mouseX, viewPort[3] - this.mouseY, 1, 1,
					viewPort);

			int w = this.width;
			int h = this.height;

			double aspect = w / h;

			glu.gluPerspective(65.0, aspect, 2.0, 60.0);
			
			gl.glMatrixMode(GL.GL_MODELVIEW);
			gl.glLoadIdentity();

			float[] eyePoint = this.camera.eyePoint;
			float[] lookAtPoint = this.camera.lookAtPoint;
			float[] upVector = this.camera.upVector;

			glu.gluLookAt(eyePoint[0], eyePoint[1], eyePoint[2], lookAtPoint[0],
					lookAtPoint[1], lookAtPoint[2], upVector[0], upVector[1],
					upVector[2]);
		}

		return true;
	}

	@Override
	public final void endScene() {
		if (this.renderMode == Renderer.SELECTION_MODE) {
			int hits = gl.glRenderMode(GL.GL_RENDER);

			int[] a = new int[OpenGLRenderer.SELECTION_BUFFER_LENGTH];
			this.selectionBuffer.get(a);

			int index = 0;
			
			index = 3;
			
			if(a[index] == 1){
				index = 6; 
			}
			
			SelectionController.getInstance().selectionOccured(a[index]);

			gl.glMatrixMode(GL.GL_PROJECTION);
			gl.glPopMatrix();

			this.selectionBuffer.clear();
			a = null;

			this.renderMode = Renderer.RENDER_MODE;
		}

	}

	@Override
	public final void clearBuffers() {
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
	}

	public final void drawElements() {

		if (ViewFrustrum.getInstance().isBoundingSphererInFrustrum(
				this.geometry.boundingVolume)) {
			Texture texture = TextureController.getInstance().getTexture(
					this.geometry.getTextureID());

			gl.glBindTexture(GL.GL_TEXTURE_2D, texture.bindID);

			
			
			if (this.renderMode == Renderer.SELECTION_MODE) {
				this.determineSelectionState();
			}

			gl.glVertexPointer(4, GL.GL_FLOAT, 0, this.geometry.vertexBuffer);
			gl.glTexCoordPointer(2, GL.GL_FLOAT, 0,
					this.geometry.textureCoordinateBuffer);
			gl.glNormalPointer(GL.GL_FLOAT, 0, this.geometry.normalBuffer);

			gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT,
					this.geometry.ambientFactor);
			gl.glMaterialfv(GL.GL_FRONT, GL.GL_DIFFUSE,
					this.geometry.diffuseFactor);
			gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR,
					this.geometry.specularFactor);
			gl.glMaterialf(GL.GL_FRONT, GL.GL_EMISSION,
					this.geometry.emmisionFactor);
			
			
			gl.glMaterialf(GL.GL_FRONT, GL.GL_SHININESS, this.geometry.shinessFactor);

			gl.glBegin(GL.GL_TRIANGLES);

			for (int i = 0; i < this.geometry.INDEX_COUNT; i++) {
				gl.glArrayElement(geometry.indices[i]);
			}

			gl.glEnd();
		}

	}

	public final void determineSelectionState() {
		if (this.geometry.isSelectable) {
			gl.glPopName();
			gl.glPushName(this.geometry.selectionID);
		} else {
			gl.glPopName();
		}

	}

	@Override
	public final void setWidth() {
		this.width = ((Component) this.drawable).getWidth();

	}

	@Override
	public final void setHeight() {
		this.height = ((Component) this.drawable).getHeight();

	}

	@Override
	public final void setBackgroundColor(float[] color) {
		gl.glClearColor(color[0], color[1], color[2], color[3]);
	}

	@Override
	public final void setClearDepth(int i) {
		gl.glClearDepth(i);
	}

	public final Texture createTexture(String name) {
		int textureID;
		int retval[] = new int[1];

		gl.glGenTextures(1, retval);
		textureID = retval[0];

		GarbageCollector.getInstance().addTextureObjectID(textureID);

		Texture texture = new Texture(name, textureID);

		gl.glBindTexture(GL.GL_TEXTURE_2D, texture.bindID);
		glu.gluBuild2DMipmaps(GL.GL_TEXTURE_2D, GL.GL_RGB8, texture.imgWidth,
				texture.imgHeight, GL.GL_RGB, GL.GL_UNSIGNED_BYTE,
				texture.textureBuffer);

		gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER,
				GL.GL_LINEAR_MIPMAP_LINEAR);
		gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER,
				GL.GL_LINEAR_MIPMAP_LINEAR);
		gl
				.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S,
						GL.GL_REPEAT);
		gl
				.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T,
						GL.GL_REPEAT);

		gl.glTexEnvi(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_MODULATE);
		
		float[] filter = new float[1];
		gl.glGetFloatv(GL.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, filter);
		gl.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, filter[0]);
		
		texture.releaseBuffer();

		return texture;

	}

	public final void deleteTextureObjects(int[] textureObjectsIDList) {
		gl.glDeleteTextures(textureObjectsIDList.length, textureObjectsIDList);
	}

	@Override
	public final void executeState() {
		// this.geometry.executeState();

	}

	@Override
	public final int[] generateBuffer() {
		int[] bufferID = new int[1];

		gl.glGenBuffers(1, bufferID);

		return bufferID;
		/*
		 * gl.glBindBuffer(GL.GL_ARRAY_BUFFER, bufferID[0]);
		 * 
		 * gl.glBufferData(GL.GL_ARRAY_BUFFER, Geometry.SIZEOF_FLOAT
		 * geometry.VERTEX_COUNT * geometry.DIMENSION, (Buffer) null,
		 * GL.GL_STREAM_DRAW);
		 */

	}

	public final void processMouseSelection(int x, int y) {
		this.mouseX = x;
		this.mouseY = y;

		this.renderMode = Renderer.SELECTION_MODE;
	}

	public final void flushBuffers() {
		// gl.glFlush();
	}

	private final static IntBuffer newIntBuffer(int numElements) {

		ByteBuffer bb = newByteBuffer(numElements * SIZEOF_INT);
		return bb.asIntBuffer();
	}

	private final static ByteBuffer newByteBuffer(int numElements) {
		ByteBuffer bb = ByteBuffer.allocateDirect(numElements);
		bb.order(ByteOrder.nativeOrder());
		return bb;
	}

	@Override
	public void setRenderMode(int mode) {
		// TODO Auto-generated method stub

	}
}
