package game.engine.spatial;


import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;


public class Transformation {

	public static final int SIZEOF_FLOAT = 4;

	public float[][] compositeMatrix = null;

	public Transformation() {
		this.compositeMatrix = new float[4][4];
		LinearAlgebra4D.setIdentity(this.compositeMatrix);
	}
	
	public Transformation(float[][] compositeMatrix){
		this.compositeMatrix = compositeMatrix.clone();
	}

	public final void setIdentity() {
		LinearAlgebra4D.setIdentity(this.compositeMatrix);
	}

	public static final void concateTransformations(Transformation transformation1,
			Transformation transformation2, Transformation transformationDest) {
		LinearAlgebra4D.calculateTableTableProduct(
				transformation2.compositeMatrix,
				transformation1.compositeMatrix,
				transformationDest.compositeMatrix);
	}

	public final void translate(float[] translateVector) {
		float[][] transTable = new float[4][4];

		LinearAlgebra4D.setIdentity(transTable);

		transTable[0][3] = translateVector[0];
		transTable[1][3] = translateVector[1];
		transTable[2][3] = translateVector[2];
		transTable[3][3] = 1;

		LinearAlgebra4D.calculateTableTableProduct(transTable, compositeMatrix,
				compositeMatrix);
	}

	public final void rotateRadian(float[] p1, float[] p2, float radianAngle) {

		float[][] matQuatRot = new float[4][4];

		float dx = (p2[0] - p1[0]);
		float dy = (p2[1] - p1[1]);
		float dz = (p2[2] - p1[2]);

		float axisVectLength = (float) Math.sqrt((dx * dx) + (dy * dy)
				+ (dz * dz));

		float cosA = (float) Math.cos(radianAngle);
		float oneC = 1 - cosA;
		float sinA = (float) Math.sin(radianAngle);

		float ux = dx / axisVectLength;
		float uy = dy / axisVectLength;
		float uz = dz / axisVectLength;

		float[] translate = { -p1[0], -p1[1], -p1[2], p1[3] };

		this.translate(translate);
		LinearAlgebra4D.setIdentity(matQuatRot);

		matQuatRot[0][0] = ux * ux * oneC + cosA;
		matQuatRot[0][1] = ux * uy * oneC - uz * sinA;
		matQuatRot[0][2] = ux * uz * oneC + uy * sinA;

		matQuatRot[1][0] = uy * ux * oneC + uz * sinA;
		matQuatRot[1][1] = uy * uy * oneC + cosA;
		matQuatRot[1][2] = uy * uz * oneC - ux * sinA;

		matQuatRot[2][0] = uz * ux * oneC - uy * sinA;
		matQuatRot[2][1] = uz * uy * oneC + ux * sinA;
		matQuatRot[2][2] = uz * uz * oneC + cosA;

		LinearAlgebra4D.calculateTableTableProduct(matQuatRot,
				this.compositeMatrix, this.compositeMatrix);

		this.translate(p1);
	}

	public final void rotateDegree(float[] p1, float[] p2, float degreeAngle) {
		float radianAngle = 0;
		radianAngle = (float) (degreeAngle * ((2 * Math.PI) / 360));

		this.rotateRadian(p1, p2, radianAngle);
	}

	public final void rotateAtPosition(float[] p2, float degreeAngle) {
		float[] translate = { -this.compositeMatrix[0][3],
				-this.compositeMatrix[1][3], -this.compositeMatrix[2][3],
				this.compositeMatrix[3][3] };

		this.translate(translate);

		float radianAngle = 0;
		radianAngle = (float) (degreeAngle * ((2 * Math.PI) / 360));

		float[] p1 = { 0, 0, 0, 1 };

		this.rotateAtOrigin(p1, p2, radianAngle);

		float[] translate2 = { -translate[0], -translate[1], -translate[2],
				translate[3] };

		this.translate(translate2);
	}

	private final void rotateAtOrigin(float[] p1, float[] p2, float radianAngle) {
		float[][] matQuatRot = new float[4][4];

		float dx = (p2[0] - p1[0]);
		float dy = (p2[1] - p1[1]);
		float dz = (p2[2] - p1[2]);

		float axisVectLength = (float) Math.sqrt((dx * dx) + (dy * dy)
				+ (dz * dz));

		float cosA = (float) Math.cos(radianAngle);
		float oneC = 1 - cosA;
		float sinA = (float) Math.sin(radianAngle);

		float ux = dx / axisVectLength;
		float uy = dy / axisVectLength;
		float uz = dz / axisVectLength;

		LinearAlgebra4D.setIdentity(matQuatRot);

		matQuatRot[0][0] = ux * ux * oneC + cosA;
		matQuatRot[0][1] = ux * uy * oneC - uz * sinA;
		matQuatRot[0][2] = ux * uz * oneC + uy * sinA;

		matQuatRot[1][0] = uy * ux * oneC + uz * sinA;
		matQuatRot[1][1] = uy * uy * oneC + cosA;
		matQuatRot[1][2] = uy * uz * oneC - ux * sinA;

		matQuatRot[2][0] = uz * ux * oneC - uy * sinA;
		matQuatRot[2][1] = uz * uy * oneC + ux * sinA;
		matQuatRot[2][2] = uz * uz * oneC + cosA;

		LinearAlgebra4D.calculateTableTableProduct(matQuatRot,
				this.compositeMatrix, this.compositeMatrix);
	}

	public final void scale(float[] scaleVector, float[] pivotPoint) {

		float[][] scaleTable = new float[4][4];

		LinearAlgebra4D.setIdentity(scaleTable);

		scaleTable[0][0] = scaleVector[0];
		scaleTable[0][3] = (1 - scaleVector[0]) * pivotPoint[0];

		scaleTable[1][1] = scaleVector[1];
		scaleTable[1][3] = (1 - scaleVector[1]) * pivotPoint[1];

		scaleTable[2][2] = scaleVector[2];
		scaleTable[2][3] = (1 - scaleVector[2]) * pivotPoint[2];

		LinearAlgebra4D.calculateTableTableProduct(scaleTable,
				this.compositeMatrix, this.compositeMatrix);
	}

	public final void scaleAtPosition(float[] scaleVector) {

		float[][] scaleTable = new float[4][4];

		LinearAlgebra4D.setIdentity(scaleTable);

		scaleTable[0][0] = scaleVector[0];
		scaleTable[0][3] = (1 - scaleVector[0]) * this.compositeMatrix[0][3];

		scaleTable[1][1] = scaleVector[1];
		scaleTable[1][3] = (1 - scaleVector[1]) * this.compositeMatrix[1][3];

		scaleTable[2][2] = scaleVector[2];
		scaleTable[2][3] = (1 - scaleVector[2]) * this.compositeMatrix[2][3];

		LinearAlgebra4D.calculateTableTableProduct(scaleTable,
				this.compositeMatrix, this.compositeMatrix);
	}

	public static void main(String args[]) {
		Transformation trans = new Transformation();
		Transformation trans2 = new Transformation();
		Transformation trans3 = new Transformation();
		
		float[] translate = {1,0,1,1};
		float[] translate2 = {-5,0,5,1};
		
		float[] point = {4,0,4,1};
		trans.translate(translate);	
		trans2.rotateDegree(LinearAlgebra4D.ORIGIN, LinearAlgebra4D.UNION_VECTOR_Y, 90);
		trans3.translate(translate2);
		
		Transformation.concateTransformations(trans, trans2,trans);
		Transformation.concateTransformations(trans, trans3, trans);
		
		LinearAlgebra4D.calculateTableVectorProduct(trans.compositeMatrix, point, point);
		
		for(int i = 0; i < 4; i++){
			System.out.println((int)point[i]);
		}
		
		//LinearAlgebra4D.displayTable(trans.compositeMatrix);
	}
}
